package Algorithm;
/**

 * This program creates an undirected graph G from an input file
 * where the edges are represented as pair of integers
 * and integers represent vertices of the graph.
 * It is assumed that there is no duplicate edge and self loop in the graph.
 * */
import java.io.*;
import java.lang.reflect.Array;
import java.util.*;

import utils.SetOperations;
import utils.Vscore;

public class Graph {
	public ArrayList <HashSet<Integer>> AdjList;
	public ArrayList <HashSet<Integer>> M ; // used in Peamc
	
	private Vscore[] S;
	private int degeneracy; // contains degeneracy value
	
	HashMap <String, Integer> vertexIndex ;
	int numberOfVertices;
	public ArrayList <Integer> vertices;

	private Map<Integer, Integer> degenscores; // (key, value) pair; key = vertex, value = degeneracy score
	private Map<Integer, Integer> trianglecount; // (key, value) pair; key = vertex, value = triangle count
	
	ArrayList<Integer> degreeOrder;
	ArrayList<Integer> triangleOrder;
	ArrayList<Integer> degenOrder;
	
	public Graph() {
		this.AdjList = new ArrayList<>();
		this.vertexIndex = new HashMap<>() ;
		this.numberOfVertices = 0 ;
		this.vertices = new ArrayList<>() ;
		degreeOrder = new ArrayList<>();
		triangleOrder = new ArrayList<>();
		degenOrder = new ArrayList<>();
		trianglecount = new HashMap<>();
	}

	public Graph(Graph G) {
		this.AdjList = new ArrayList<>(G.AdjList);
		this.vertexIndex = new HashMap<>() ;
		this.numberOfVertices = 0 ;
		this.vertices = new ArrayList<>() ;
		degreeOrder = new ArrayList<>();
		triangleOrder = new ArrayList<>();
		degenOrder = new ArrayList<>();
		trianglecount = new HashMap<>();
	}

	public void buildM () {
		M = new ArrayList<>() ;
		for (Integer vertex : this.vertices) {
			M.add(new HashSet<>()) ;
			for (Integer neighbor : AdjList.get(vertex)) {
				if (neighbor > vertex)
					M.get(vertex).add(neighbor) ;
			}
		}
	}

	public boolean isAllDigits (String s) {
		for (int i = 0; i < s.length(); i++) {
			if (Character.isDigit(s.charAt(i)) == false) return false ;
		}
		return true ;
	}

	public int addVertex (String vertex) {
		if (!vertexIndex.containsKey(vertex)) {
			this.vertices.add(numberOfVertices) ;
			degreeOrder.add(numberOfVertices);
			degenOrder.add(numberOfVertices);
			triangleOrder.add(numberOfVertices);
			this.vertexIndex.put(vertex, numberOfVertices++) ;
			this.AdjList.add(new HashSet<>()) ;
		}
		return vertexIndex.get(vertex) ;
	}

	public Graph(String args) {
		this.AdjList = new ArrayList<>();
		this.vertexIndex = new HashMap<>() ;
		this.numberOfVertices = 0 ;
		this.vertices = new ArrayList<>() ;
		degreeOrder = new ArrayList<>();
		triangleOrder = new ArrayList<>();
		degenOrder = new ArrayList<>();
		trianglecount = new HashMap<>();
		try {
			BufferedReader br = new BufferedReader(new FileReader(args));
			String line;
			while ((line = br.readLine()) != null) {
				String[] splits = line.split("\\s+");
				if (splits.length < 2) continue ;
				if (isAllDigits(splits [0]) == false || isAllDigits(splits [1]) == false) continue ;

				addEdge(splits[0],splits[1]);
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		Collections.sort(degreeOrder, (a,b) -> this.degreeOf(a) - this.degreeOf(b));
		Collections.sort(triangleOrder, (a,b) -> this.tcountAt(a) - this.tcountAt(b));
	}

	public void addEdge(String x, String y) {
		
		int u = addVertex(x);
		int v = addVertex(y);
		
		if (u == v) return ;
		if (this.AdjList.get(v).contains(u) == false) {

			if (!trianglecount.keySet().contains(u))
				trianglecount.put(u, 0);
			if (!trianglecount.keySet().contains(v))
				trianglecount.put(v, 0);

			this.AdjList.get(v).add(u) ;
			this.AdjList.get(u).add(v) ;

			// update the triangles count
			HashSet<Integer> commonNeighborhood = SetOperations.intersect(AdjList.get(u), AdjList.get(v));

			for (int w : commonNeighborhood) {

				int tcount_u = 0;
				int tcount_v = 0;
				int tcount_w = 0;

				tcount_u = trianglecount.get(u);
				tcount_v = trianglecount.get(v);
				tcount_w = trianglecount.get(w);

				tcount_u++;
				tcount_v++;
				tcount_w++;

				trianglecount.put(u, tcount_u);
				trianglecount.put(v, tcount_v);
				trianglecount.put(w, tcount_w);
			}
		}
	}
	
	public boolean containsEdge(int u, int v){
	    if(AdjList.get(u) != null){
	        if(AdjList.get(u).contains(v))
	            return true;
	        else
	            return false;
	    }
	    return false;
	}
	


	public void removeEdge(int u, int v) {

		if (AdjList.get(u) != null) {
			if (AdjList.get(u).contains(v)) {
				AdjList.get(u).remove(v);
				AdjList.get(v).remove(u);
			}
		}

	}

	public void print() {
		for (int v = 0 ; v < numberOfVertices ; v ++){
			System.out.print(v + ":");
			for (int u : AdjList.get(v)) {
				System.out.print(u + ";");
			}
			System.out.println();
		}
		System.out.println();
	}

	public void print(FileWriter fw) {
		try {
			for (int v = 0 ; v < numberOfVertices ; v ++){
				fw.write(v + " ");
				for (int u : AdjList.get(v)) {
					fw.write(u + " ");
				}
				fw.write("\n");
			}
		} catch (IOException e) {
		}
	}

	public int MaxDegree() {
		int max = Integer.MIN_VALUE;
		for (int v = 0 ; v < numberOfVertices ; v ++){
			int size = AdjList.get(v).size();
			if (max < size) {
				max = size;
			}
		}
		return max;
	}

	public int MinDegree() {
		int min = Integer.MAX_VALUE;
		for (int v = 0 ; v < numberOfVertices ; v ++){
			int size = AdjList.get(v).size();
			if (min > size) {
				min = size;
			}
		}
		return min;
	}

	public int AvgDegree() {
		int sum = 0;
		for (int v = 0 ; v < numberOfVertices ; v ++){
			sum += AdjList.get(v).size();
		}
		return (sum / (2 * AdjList.size()));
	}

	public int numE() {
		int size = 0;
		for (int v = 0 ; v < numberOfVertices ; v ++){
			size += AdjList.get(v).size();
		}
		return size / 2;
	}

	public int numV() {
		return this.numberOfVertices ;
	}

	public int getSize() {
		int size = 0;
		for (int v = 0 ; v < numberOfVertices ; v ++){
			size++;
			size += AdjList.get(v).size();
		}
		return size;
	}

	public ArrayList<Integer> getDegenOrdering() {
		return degenOrder;
	}
	
	public ArrayList<Integer> getDegreeOrdering(){
		return degreeOrder;
	}
	
	public ArrayList<Integer> getTriangleOrdering(){
		return triangleOrder;
	}

	public int degreeOf(int vertex) {
		return this.AdjList.get(vertex).size();
	}

	public int degeneracyOf(int vertex) {
		return degenscores.get(vertex);
	}

	public int tcountAt(int vertex) {
		return trianglecount.get(vertex);
	}

	public Collection<Integer> neighborsOf(int u) {
		return this.AdjList.get(u);
	}

	public void clear() {
		AdjList.clear();
	}

	// Adopted this code from
	// https://github.com/jgrapht/jgrapht/blob/master/jgrapht-core/src/main/java/org/jgrapht/alg/scoring/Coreness.java
	public void computeDegeneracy() {

		if (degenscores != null) {
			return;
		}

		degenscores = new HashMap<>();

		S = new Vscore[AdjList.size()];

		int vcount = 0;

		degeneracy = 0;

		/* initialize buckets */
		int n = AdjList.size();
		int maxDegree = n - 1;
		Set<Integer>[] buckets = (Set<Integer>[]) Array.newInstance(Set.class, maxDegree + 1);

		for (int i = 0; i < buckets.length; i++) {
			buckets[i] = new HashSet<>();
		}

		int minDegree = n;
		Map<Integer, Integer> degrees = new HashMap<>();

		for (int v = 0 ; v < numberOfVertices ; v ++){
			int d = degreeOf(v);
			buckets[d].add(v);
			degrees.put(v, d);
			minDegree = Math.min(minDegree, d);
		}

		/* Extract from bucket */
		while (minDegree < n) {
			Set<Integer> b = buckets[minDegree];
			if (b.isEmpty()) {
				minDegree++;
				continue;
			}

			int v = b.iterator().next();
			b.remove(v);

			degenscores.put(v, minDegree);
			S[vcount] = new Vscore(v, minDegree);
			vcount++;

			degeneracy = Math.max(degeneracy, minDegree);

			for (int u : AdjList.get(v)) {
				int uDegree = degrees.get(u);
				if (uDegree > minDegree && !degenscores.containsKey(u)) {
					buckets[uDegree].remove(u);
					uDegree--;
					degrees.put(u, uDegree);
					buckets[uDegree].add(u);
					minDegree = Math.min(minDegree, uDegree);
				}
			}
		}
		//Arrays.parallelSort(S);
		Collections.sort(degenOrder, (a,b) -> this.degeneracyOf(a) - this.degeneracyOf(b));
	}

	public static void main(String[] args) {
		// TODO Auto-generated method stub

		Graph g = new Graph("C:/Users/apurd/Documents/data/sx_stackoverflow_a2q_edges");

		int maxdegree = g.MaxDegree();

		System.out.println("max degree of sx_stackoverflow: " + maxdegree);

		// g = new Graph("/home/apurba/data/ER-2M-15M-edges");

		// maxdegree = g.MaxDegree();

		// System.out.println("max degree of 2M-15M: " + maxdegree);

		// g.removeEdge(1,2);
		// g.removeEdge(2,3);
		// System.out.println("After removal of (1,2) and (2,3)");
		// g.print();

		// new Tomita(g);

	}

	public void printedges(FileWriter fw) throws Exception {
		String[] edges = new String[numE()];

		int count_edges = 0;
		for (int v = 0 ; v < numberOfVertices ; v ++){
			for (int u : AdjList.get(v)) {
				if (u < v) {
					edges[count_edges] = u + " " + v;
					count_edges++;
				}
			}
		}

		System.out.println(count_edges);

		// randomly shuffling the edges
		for (int i = 0; i < edges.length; i++) {
			int r = i + (int) (Math.random() * (edges.length - i));
			String temp = edges[r];
			edges[r] = edges[i];
			edges[i] = temp;
		}

		for (int i = 0; i < edges.length; i++) {
			fw.write(edges[i] + "\n");
		}
		fw.close();

	}

	public void printVertices(FileWriter fw) throws Exception {
		for (int v = 0 ; v < numberOfVertices ; v ++){
			fw.write(v + "\n");
		}

	}

	public int getId(int v) {
		
		return this.vertexIndex.get(String.valueOf(v));
	}



}
